//open source
var http = require("http");
var https = require('https');
const fs = require('fs');
var connect = require('connect');
var winston = require('winston');
var path = require('path');

//arguments are <evo-node> <js file> <uid> <gid> <ppid> <config file>
var config = require(process.argv[5]);
var app = connect();

//ours; js
var rotateSize = require('./fileRotateSize');
var hlp = require('./helper');
var authproxy = require('./req_handlers/authproxy');
var phpreq = require('./req_handlers/php');
var httpstreamreq = require('./req_handlers/httpstream');
var defreq = require('./req_handlers/default');
var resphdrs = require('./req_handlers/resphdrs');
var bridge = null;

//ours: c++
var ews = require('./ews.node');

if (config.loggers.file.filename == "") {
	config.loggers.file.filename = "evo-webserver";
}

if (config.loggers.file.dirname == "") {
	config.loggers.file.dirname = "../logs";
}

var binDir = path.dirname(require.main.filename);
var isWin = /^win/.test(process.platform);
if (isWin) {
	config.loggers.file.dirname = path.resolve(binDir, 
		config.loggers.file.dirname);
} else {
	config.loggers.file.dirname = path.resolve(binDir, '..',
		config.loggers.file.dirname);
}

config.loggers.file.handleExceptions = true;

var logConsole = new (winston.transports.Console)(config.loggers.console);
var logFile = new (winston.transports.FileRotateSize)(config.loggers.file);
logConsole.timestamp = logFile.timestamp = function() {
	return new Date().getTime() + ':' + process.pid;
}
logFile.formatter = function(options) {
	// Return string will be passed to logger.
	return options.timestamp() +' '+ options.level +
		' '+ (undefined !== options.message ? options.message : '') + 
		(options.meta && Object.keys(options.meta).length ? '\n\t'+ 
		JSON.stringify(options.meta) : '' );
}

function loggerError(err) {
  	console.log('warning: A logger error was encountered: ' + err);
  	console.log('warning: node-ews will continue without file logging.');
  	return false;
}

var logger = new (winston.Logger)({
    transports: [
		logConsole, logFile
    ],
    exitOnError: loggerError
});

function logMessage(level, msg) {
	try {
		logger.log(level, msg);
	} catch (err) {

	}
}

logMessage('info', 'nodeEWS started.');
logMessage('info', 'Logging to ' + config.loggers.file.dirname + ".");

function messageLogger(level, file, line, func, msg) {
	var message = file + ':' + line + ':' + func + ':' + msg;
	logMessage(level, message);
}

function exit() {
	logMessage('info', 'nodeEWS quitted.');
	server.close();
	process.exit();
}

function createHttpsServer() {
	try {
		config.sslKeyFile = path.resolve(binDir, config.sslKeyFile);
		config.sslCertFile = path.resolve(binDir, config.sslCertFile);
		var options = {
			key: fs.readFileSync(config.sslKeyFile),
		  	cert: fs.readFileSync(config.sslCertFile)
		};
		return https.createServer(options, app);
	} catch (err) {
		logMessage('error', err);
		return null;
	}
}

var server = null;
var ishttps = false;
if (config.sslMode == "enabled") {
	server = createHttpsServer();
	if (server) {
		ishttps = true;
	}
} else if (config.sslMode == "auto") {
	server = createHttpsServer();
	if (server == null) {
		logMessage('info', 'Reverting to HTTP.');
		server = http.createServer(app);
	} else {
		ishttps = true;
	}
} else { //disabled
	server = http.createServer(app);
}
server.on('error', (e) => {
	logMessage('error', 'Could not create the http(s) server.', 'Error:', 
		e.message, '. Quitting now!');
	exit();	
});

server.on('connection', (socket) => {
	if (socket.remoteAddress != undefined) {
		var ipv4 = hlp.formatIP(socket.remoteAddress);
		socket.on('close', () => {
			logMessage('debug', '[' + ipv4 + ':' + socket.remotePort + 
				'] disconnected');
		});
		if (bridge != null && bridge.isIPFilterEnabled()) {
			bridge.tryAllowIP(ipv4);
			if (bridge.isIPAllowed(ipv4)) {
				logMessage('debug', '[' + ipv4 + ':' + socket.remotePort + 
					'] connected');	
			} else {
				logMessage('debug', '[' + ipv4 + ':' + socket.remotePort + 
					'] blocked');
				socket.end();
			}
		} else {
			logMessage('debug', '[' + ipv4 + ':' + socket.remotePort + 
				'] connected');
		}
	}
});

server.on('listening', () => {
	logMessage('info', (ishttps ? 'HTTPS' : 'HTTP') + 
		' server is listening on port ' + config.port + '.');

	bridge = new ews.JSBridge(messageLogger, config.webRootFolder, 
	config.hasGroupNameAliases, config.emsPort, 
	config.mediaFileDownloadTimeout, parseInt(process.argv[4]));

	/*
		Load supported mime types.
	*/
	for (var i=0; i<config.supportedMimeTypes.length;++i) {
		var mt = config.supportedMimeTypes[i];
		var exts = mt.extensions.trim().split(',');
		for (var j=0;j<exts.length; ++j) {
			bridge.addSupportedMimeType(exts[j], mt.mimeType, mt.streamType, 
				mt.isManifest);
			if (!mt.isManifest) {
				bridge.addSupportedNonStreamingMediaFile(exts[j]);
			}
		}	
	}
	logMessage('info', 'Supported mime types are loaded.');

	if (config.enableIPFilter) {
		config.whitelistFile = path.resolve(binDir, config.whitelistFile);
		config.blacklistFile = path.resolve(binDir, config.blacklistFile);		
		bridge.loadIPFilter(config.whitelistFile, config.blacklistFile);
	}
	logMessage('info', 'IP Filtering is ' + (bridge.isIPFilterEnabled() ? 
		'enabled' : 'disabled') + '.');

	bridge.asyncSpinIOHandler(() => {
		exit();
	});

	//setup the middlewares
	//authentication proxy handler
	if (config.apiProxy.enable) {
		if (config.apiProxy.authentication != "basic") {
			logMessage('error', 
				'Only basic authentication is supported. Check configuration.');
			return;
		} else if (config.apiProxy.pseudoDomain == "" || 
				config.apiProxy.address == "" ||
				config.apiProxy.port == "" ||
				config.apiProxy.userName == "" ||
				config.apiProxy.password == "") {
			logMessage('error', 
				'Domain, address, port, username or password cannot be blank.',
				'Check configuration.');
			return;
		}
		app.use(authproxy.handle(config.apiProxy, logger));
	}
	//response header decorator
	app.use(resphdrs.handle(config.includeResponseHeaders, logger));
	//php request handler
	app.use(phpreq.handle(logger, config));
	//ranged http streaming request handler
	app.use(httpstreamreq.handle(bridge, logger, config));
	//default request handler
	app.use(defreq.handle(logger, config));
});

server.listen(config.port, config.bindToIP);

